Imports System

Namespace BackPropagation
  Friend Class BackPropagationProgram
	Shared Sub Main(ByVal args() As String)
	  Try
		Console.WriteLine(vbCrLf & "Begin Neural Network Back-Propagation demo" & vbCrLf)

		Console.WriteLine("Creating a 3-input, 4-hidden, 2-output neural network")
		Console.WriteLine("Using sigmoid function for input-to-hidden activation")
		Console.WriteLine("Using tanh function for hidden-to-output activation")
		Dim nn As New NeuralNetwork(3, 4, 2)

		' arbitrary weights and biases
		Dim weights() As Double = { 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, -2.0, -6.0, -1.0, -7.0, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, -2.5, -5.0 }

		Console.WriteLine(vbCrLf & "Initial 26 random weights and biases are:")
		Helpers.ShowVector(weights, 2, True)

		Console.WriteLine("Loading neural network weights and biases")
		nn.SetWeights(weights)

		Console.WriteLine(vbCrLf & "Setting inputs:")
		Dim xValues() As Double = { 1.0, 2.0, 3.0 }
		Helpers.ShowVector(xValues, 2, True)

		Dim initialOutputs() As Double = nn.ComputeOutputs(xValues)
		Console.WriteLine("Initial outputs:")
		Helpers.ShowVector(initialOutputs, 4, True)

		Dim tValues() As Double = { -0.8500, 0.7500 } ' target (desired) values. note these only make sense for tanh output activation
		Console.WriteLine("Target outputs to learn are:")
		Helpers.ShowVector(tValues, 4, True)

		Dim eta As Double = 0.90 ' learning rate - controls the maginitude of the increase in the change in weights. found by trial and error.
		Dim alpha As Double = 0.04 ' momentum - to discourage oscillation. found by trial and error.
		Console.WriteLine("Setting learning rate (eta) = " & eta.ToString("F2") & " and momentum (alpha) = " & alpha.ToString("F2"))

		Console.WriteLine(vbCrLf & "Entering main back-propagation compute-update cycle")
		Console.WriteLine("Stopping when sum absolute error <= 0.01 or 1,000 iterations" & vbCrLf)
		Dim ctr As Integer = 0
		Dim yValues() As Double = nn.ComputeOutputs(xValues) ' prime the back-propagation loop
		Dim [error] As Double = BackPropagationProgram.Error(tValues, yValues)
		Do While ctr < 1000 AndAlso [error] > 0.01
		  Console.WriteLine("===================================================")
		  Console.WriteLine("iteration = " & ctr)
		  Console.WriteLine("Updating weights and biases using back-propagation")
		  nn.UpdateWeights(tValues, eta, alpha)
		  Console.WriteLine("Computing new outputs:")
		  yValues = nn.ComputeOutputs(xValues)
		  Helpers.ShowVector(yValues, 4, False)
		  Console.WriteLine(vbCrLf & "Computing new error")
		  [error] = BackPropagationProgram.Error(tValues, yValues)
		  Console.WriteLine("Error = " & [error].ToString("F4"))
		  ctr += 1
		Loop
		Console.WriteLine("===================================================")
		Console.WriteLine(vbCrLf & "Best weights and biases found:")
                Dim bestWeights() As Double = nn.GetWeights
		Helpers.ShowVector(bestWeights, 2, True)

		Console.WriteLine("End Neural Network Back-Propagation demo" & vbCrLf)
		Console.ReadLine()
	  Catch ex As Exception
		Console.WriteLine("Fatal: " & ex.Message)
		Console.ReadLine()
	  End Try
	End Sub ' Main

	Private Shared Function [Error](ByVal target() As Double, ByVal output() As Double) As Double ' sum absolute error. could put into NeuralNetwork class.
	  Dim sum As Double = 0.0
	  For i  = 0 To target.Length - 1
		sum += Math.Abs(target(i) - output(i))
	  Next i
	  Return sum
	End Function
  End Class ' class BackPropagation

  Friend Class NeuralNetwork
	Private numInput As Integer
	Private numHidden As Integer
	Private numOutput As Integer

	Private inputs() As Double
        Private ihWeights()() As Double ' input-to-hidden
	Private ihSums() As Double
	Private ihBiases() As Double
	Private ihOutputs() As Double

        Private hoWeights()() As Double ' hidden-to-output
	Private hoSums() As Double
	Private hoBiases() As Double
	Private outputs() As Double

	Private oGrads() As Double ' output gradients for back-propagation
	Private hGrads() As Double ' hidden gradients for back-propagation

        Private ihPrevWeightsDelta()() As Double ' for momentum with back-propagation
	Private ihPrevBiasesDelta() As Double

        Private hoPrevWeightsDelta()() As Double
	Private hoPrevBiasesDelta() As Double

	Public Sub New(ByVal numInput As Integer, ByVal numHidden As Integer, ByVal numOutput As Integer)
	  Me.numInput = numInput
	  Me.numHidden = numHidden
	  Me.numOutput = numOutput

            inputs = New Double(numInput - 1) {}
            ihWeights = Helpers.MakeMatrix(numInput, numHidden)
	  ihSums = New Double(numHidden - 1){}
	  ihBiases = New Double(numHidden - 1){}
	  ihOutputs = New Double(numHidden - 1){}
	  hoWeights = Helpers.MakeMatrix(numHidden, numOutput)
	  hoSums = New Double(numOutput - 1){}
	  hoBiases = New Double(numOutput - 1){}
	  outputs = New Double(numOutput - 1){}

	  oGrads = New Double(numOutput - 1){}
	  hGrads = New Double(numHidden - 1){}

	  ihPrevWeightsDelta = Helpers.MakeMatrix(numInput, numHidden)
	  ihPrevBiasesDelta = New Double(numHidden - 1){}
	  hoPrevWeightsDelta = Helpers.MakeMatrix(numHidden, numOutput)
	  hoPrevBiasesDelta = New Double(numOutput - 1){}
	End Sub

	Public Sub UpdateWeights(ByVal tValues() As Double, ByVal eta As Double, ByVal alpha As Double) ' update the weights and biases using back-propagation, with target values, eta (learning rate), alpha (momentum)
	  ' assumes that SetWeights and ComputeOutputs have been called and so all the internal arrays and matrices have values (other than 0.0)
	  If tValues.Length <> numOutput Then
		Throw New Exception("target values not same Length as output in UpdateWeights")
	  End If

	  ' 1. compute output gradients
	  For i  = 0 To oGrads.Length - 1
		Dim derivative As Double = (1 - outputs(i)) * (1 + outputs(i)) ' derivative of tanh
		oGrads(i) = derivative * (tValues(i) - outputs(i))
	  Next i

	  ' 2. compute hidden gradients
	  For i  = 0 To hGrads.Length - 1
		Dim derivative As Double = (1 - ihOutputs(i)) * ihOutputs(i) ' (1 / 1 + exp(-x))' -- using output value of neuron
		Dim sum As Double = 0.0
		For j  = 0 To numOutput - 1 ' each hidden delta is the sum of numOutput terms
		  sum += oGrads(j) * hoWeights(i)(j) ' each downstream gradient * outgoing weight
		Next j
		hGrads(i) = derivative * sum
	  Next i

	  ' 3. update input to hidden weights (gradients must be computed right-to-left but weights can be updated in any order
	  For i  = 0 To ihWeights.Length - 1 ' 0..2 (3)
		For j  = 0 To ihWeights(0).Length - 1 ' 0..3 (4)
		  Dim delta As Double = eta * hGrads(j) * inputs(i) ' compute the new delta
		  ihWeights(i)(j) += delta ' update
		  ihWeights(i)(j) += alpha * ihPrevWeightsDelta(i)(j) ' add momentum using previous delta. on first pass old value will be 0.0 but that's OK.
		Next j
	  Next i

	  ' 3b. update input to hidden biases
	  For i  = 0 To ihBiases.Length - 1
		Dim delta As Double = eta * hGrads(i) * 1.0 ' the 1.0 is the constant input for any bias; could leave out
		ihBiases(i) += delta
		ihBiases(i) += alpha * ihPrevBiasesDelta(i)
	  Next i

	  ' 4. update hidden to output weights
	  For i  = 0 To hoWeights.Length - 1 ' 0..3 (4)
		For j  = 0 To hoWeights(0).Length - 1 ' 0..1 (2)
		  Dim delta As Double = eta * oGrads(j) * ihOutputs(i) ' see above: ihOutputs are inputs to next layer
		  hoWeights(i)(j) += delta
		  hoWeights(i)(j) += alpha * hoPrevWeightsDelta(i)(j)
		  hoPrevWeightsDelta(i)(j) = delta
		Next j
	  Next i

	  ' 4b. update hidden to output biases
	  For i  = 0 To hoBiases.Length - 1
		Dim delta As Double = eta * oGrads(i) * 1.0
		hoBiases(i) += delta
		hoBiases(i) += alpha * hoPrevBiasesDelta(i)
		hoPrevBiasesDelta(i) = delta
	  Next i
	End Sub ' UpdateWeights

	Public Sub SetWeights(ByVal weights() As Double)
	  ' copy weights and biases in weights[] array to i-h weights, i-h biases, h-o weights, h-o biases
	  Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
	  If weights.Length <> numWeights Then
		Throw New Exception("The weights array length: " & weights.Length & " does not match the total number of weights and biases: " & numWeights)
	  End If

	  Dim k As Integer = 0 ' points into weights param

	  For i  = 0 To numInput - 1
		For j  = 0 To numHidden - 1
		  ihWeights(i)(j) = weights(k)
		  k += 1
		Next j
	  Next i

	  For i  = 0 To numHidden - 1
		ihBiases(i) = weights(k)
		k += 1
	  Next i

	  For i  = 0 To numHidden - 1
		For j  = 0 To numOutput - 1
		  hoWeights(i)(j) = weights(k)
		  k += 1
		Next j
	  Next i

	  For i  = 0 To numOutput - 1
		hoBiases(i) = weights(k)
		k += 1
	  Next i
	End Sub

	Public Function GetWeights() As Double()
	  Dim numWeights As Integer = (numInput * numHidden) + (numHidden * numOutput) + numHidden + numOutput
	  Dim result(numWeights - 1) As Double
	  Dim k As Integer = 0
	  For i  = 0 To ihWeights.Length - 1
		For j  = 0 To ihWeights(0).Length - 1
		  result(k) = ihWeights(i)(j)
		  k += 1
		Next j
	  Next i
	  For i  = 0 To ihBiases.Length - 1
		result(k) = ihBiases(i)
		k += 1
	  Next i
	  For i  = 0 To hoWeights.Length - 1
		For j  = 0 To hoWeights(0).Length - 1
		  result(k) = hoWeights(i)(j)
		  k += 1
		Next j
	  Next i
	  For i  = 0 To hoBiases.Length - 1
		result(k) = hoBiases(i)
		k += 1
	  Next i
	  Return result
	End Function

	Public Function ComputeOutputs(ByVal xValues() As Double) As Double()
	  If xValues.Length <> numInput Then
		Throw New Exception("Inputs array length " & inputs.Length & " does not match NN numInput value " & numInput)
	  End If

	  For i  = 0 To numHidden - 1
		ihSums(i) = 0.0
	  Next i
	  For i  = 0 To numOutput - 1
		hoSums(i) = 0.0
	  Next i

	  For i  = 0 To xValues.Length - 1 ' copy x-values to inputs
		Me.inputs(i) = xValues(i)
	  Next i

	  For j  = 0 To numHidden - 1 ' compute input-to-hidden weighted sums
		For i  = 0 To numInput - 1
		  ihSums(j) += Me.inputs(i) * ihWeights(i)(j)
		Next i
	  Next j

	  For i  = 0 To numHidden - 1 ' add biases to input-to-hidden sums
		ihSums(i) += ihBiases(i)
	  Next i

	  For i  = 0 To numHidden - 1 ' determine input-to-hidden output
		ihOutputs(i) = SigmoidFunction(ihSums(i))
	  Next i

	  For j  = 0 To numOutput - 1 ' compute hidden-to-output weighted sums
		For i  = 0 To numHidden - 1
		  hoSums(j) += ihOutputs(i) * hoWeights(i)(j)
		Next i
	  Next j

	  For i  = 0 To numOutput - 1 ' add biases to input-to-hidden sums
		hoSums(i) += hoBiases(i)
	  Next i

	  For i  = 0 To numOutput - 1 ' determine hidden-to-output result
		Me.outputs(i) = HyperTanFunction(hoSums(i))
	  Next i

	  Dim result(numOutput - 1) As Double ' could define a GetOutputs method instead
	  Me.outputs.CopyTo(result, 0)

	  Return result
	End Function ' ComputeOutputs

	Private Shared Function StepFunction(ByVal x As Double) As Double ' an activation function that isn't compatible with back-propagation bcause it isn't differentiable
	  If x > 0.0 Then
		  Return 1.0
	  Else
		  Return 0.0
	  End If
	End Function

	Private Shared Function SigmoidFunction(ByVal x As Double) As Double
	  If x < -45.0 Then
		  Return 0.0
	  ElseIf x > 45.0 Then
		  Return 1.0
	  Else
		  Return 1.0 / (1.0 + Math.Exp(-x))
	  End If
	End Function

	Private Shared Function HyperTanFunction(ByVal x As Double) As Double
	  If x < -10.0 Then
		  Return -1.0
	  ElseIf x > 10.0 Then
		  Return 1.0
	  Else
		  Return Math.Tanh(x)
	  End If
	End Function
  End Class ' class NeuralNetwork

  ' ===========================================================================

  Public Class Helpers
        Public Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

	Public Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal blankLine As Boolean)
	  For i  = 0 To vector.Length - 1
		If i > 0 AndAlso i Mod 12 = 0 Then ' max of 12 values per row
		  Console.WriteLine("")
		End If
		If vector(i) >= 0.0 Then
			Console.Write(" ")
		End If
		Console.Write(vector(i).ToString("F" & decimals) & " ") ' n decimals
	  Next i
	  If blankLine Then
		  Console.WriteLine(vbCrLf)
	  End If
	End Sub

        Public Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal decimals As Integer)
            Dim ct As Integer = 0
            If numRows = -1 Then ' if numRows == -1, show all rows
                numRows = Integer.MaxValue
            End If
            Dim i As Integer = 0
            Do While i < matrix.Length AndAlso ct < numRows
                For j = 0 To matrix(0).Length - 1
                    If matrix(i)(j) >= 0.0 Then ' blank space instead of '+' sign
                        Console.Write(" ")
                    End If
                    Console.Write(matrix(i)(j).ToString("F" & decimals) & " ")
                Next j
                Console.WriteLine("")
                ct += 1
                i += 1
            Loop
            Console.WriteLine("")
        End Sub
  End Class ' class Helpers
End Namespace ' ns
